require 'fileutils'

# Abstract base class for all trasfer methods
#
# @author Michal Kimle
# @abstract
# @attr_reader [String] destination path to destination directory for the transfer method
class Nifty::TransferMethod
  attr_reader :destination, :options

  class << self
    # Helper method to recognize NIFTY transfer method
    #
    # @return [TrueClass, FalseClass] whether or not class is a NIFTY transfer method
    def transfer_method?
      false
    end

    # Returns textual description of the transfer method
    # Used in help messages.
    #
    # @abstract
    # @return [String, nil] textual description of the transfer method
    def description
      nil
    end

    # Returns backend class supported by the transfer method
    #
    # @abstract
    # @return [Nifty::Backend] backend class supported by the transfer method
    def backend
      nil
    end
  end

  # Constructor
  #
  # @param [String] destination for the transfer method
  def initialize(destination, options={})
    @destination = destination
    @options = options
  end

  # Transfers file to the destination
  #
  # @param [String] file to be transfered
  # @return [String] path to the transfered file
  def transfer(file)
    fail Nifty::Errors::TransferMethods::ImageFileNotReadableError, "Image file #{file} is not readable" unless File.readable?(file)
  end

  # Removes original file before transfer method
  #
  # @param [String] file to be cleaned
  def clean_original(file)
    logger.debug("Deleting files #{file.inspect}")
    FileUtils.rm file
  rescue SystemCallError => ex
    logger.warn("Cannot delete file #{file.inspect}, #{ex.message}")
  end

  # Removes file copy if any created by transfer method
  #
  # @abstract
  # @param [String] file to be cleaned
  def clean_copy(file)
    nil
  end
end
